#!/usr/bin/env python3

import sys
import os
sys.path.append( os.path.join(os.path.dirname(__file__), '../src') )

def test_verilated_dpi_counter():
    from pueda.edalize import verilator
    verilator(
        simname='counter', top='counter',                        
        src_dir = ['./counter/rtl','./counter/verilator'], 
        inc_dir = [],
        dump_en = False, sim_en=False)

if __name__ == '__main__':
    test_verilated_dpi_counter()            


#!/usr/bin/env python3

import sys
import os
sys.path.append( os.path.join(os.path.dirname(__file__), '../src') )

import unittest

class TestStringMethods(unittest.TestCase):

    def test_edalize_icarus_hello(self):
        from pueda.edalize import icarus        
        icarus(simname='hello', top='hello_tb', src_dirs = ['./hello'], dump_en=False)

    def test_edalize_icarus_counter(self):
       from pueda.edalize import icarus
       icarus(simname='counter', top='counter_tb',
              src_dirs = ['./counter/rtl','./counter/tb'],
              iverilog_options=['-g2005-sv'],
              dump_en=False)
       # self.assertEqual('foo'.upper(), 'FOO')

    def test_pyverilator_counter(self):
        from pueda.pyverilator import pyverilator_wrapper
        pv = pyverilator_wrapper(fname='./counter/rtl/counter.v',
                                 src_dirs=['./counter/rtl'],
                                 command_args = [],
                                 dump_en = True, dump_level=1)

        for _ in range(16):
            pv.sim.clock.tick()
            print('count = %d' % pv.sim.io.count)
        assert pv.sim.io.count == 0
        pv.view_waves(mode='vcdterm', block_en=False)

    def test_verilator_counter(self):
        from test_verilator_counter import test_verilator_counter
        test_verilator_counter()

    def test_yosys_counter(self):
        from pueda.yosys import yosys
        yosys(top='counter', src_dirs = ['./counter/rtl'], synth_en=True)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3

import os
from pueda.common import get_source_files_alldir, get_inc_list, get_clean_work, write_file_lines

def yosys(top='', src_dirs = [], inc_dirs = [], exclude_files = [], synth_en=True) -> None:

    tool = 'yosys'
    work_root = get_clean_work(tool,True)

    # create yosys script
    lines  = get_inc_list(inc_dirs,prefix='read -incdir ')
    lines += ['read_verilog ' + s for s in get_source_files_alldir(src_dirs,fmts=['.v'], excludes=exclude_files)]

    lines.append('hierarchy -top %s' % top)
    fullfile = '%s_full.v' % os.path.join(work_root,top)
    lines.append('write_verilog %s' % fullfile)

    if synth_en:
        if True:
            lines += ['synth']
            lines += ['abc -g cmos2','stat']
        else:
            lines += ['synth_ice40']
        synthfile = '%s_synth.v' % os.path.join(work_root,top)
        lines.append('write_verilog %s' % synthfile)
        # lines += ['show -prefix ./ecdsa256 -format svg -viewer ']
        # lines += ['stat -tech xilinx']
    else:
        synthfile = ''

    # print output
    ysoutfile = os.path.join(work_root,'%s.ys' % top)
    write_file_lines(ysoutfile,lines,print_en=True)    

    # yosys command
    cmdstr = 'yosys -s %s > %s/yosys.log' % (ysoutfile,work_root)
    print(cmdstr)
    os.system(cmdstr)

    return {'single':fullfile,'synth':synthfile,'work':work_root}


#!/usr/bin/env python3

import os
import sys
sys.path.append( os.path.join(os.path.dirname(__file__), '..') )

from pueda.common import get_source_files_alldir, get_remote_files, get_work_path, work_exists, get_clean_work, list2str, get_inc_list

def vpi_make(src_dirs=['./'], inc_dirs=[], args = []):
    # print(src_dirs)
    src = get_source_files_alldir(src_dirs,fmts=['.c','.cc','.cpp'])
    inc = get_inc_list(inc_dirs)
    # print(inc)

    argstr = list2str(args)
    srcstr = list2str(src)
    incstr = list2str(inc)
    cmdstr = list2str( ['iverilog-vpi', argstr, srcstr, incstr]) # , ' > iverilog-vpi.log' ] )

    print('### Making VPI...')
    print(cmdstr)
    os.system(cmdstr)
    os.system('rm *.o')
    print('### Making VPI... Done')
    pass

class custom_vpi(object):
    work = './'

    def __init__(self,tool='custom_vpi',url='',flist=[],inc_dirs=[],args=[], custom_make_cmds=[],rm_en=True):
        #if not( self._init_work(tool=tool,rm_en=rm_en) ):
        self._init_work(tool=tool,rm_en=rm_en)
        self._get_custom_vpi(url=url,flist=flist)      

        if len(custom_make_cmds)>0:
            self.custom_make(cmds=custom_make_cmds)
        else:
            self.auto_make(inc_dirs=inc_dirs,args=args)
        #else:
        #    print('WARNING: skip compiling VPI %s, because folder already exists!!! ' % tool)

    def _init_work(self,tool='',rm_en=False):
        exist = work_exists(tool)

        if rm_en or not(exist):
            self.work = get_clean_work(tool=tool,rm_en=rm_en) # do not remove if exist
        else:
            self.work = get_work_path(tool)        
        
        return exist

    def _get_custom_vpi(self,url,flist):
        get_remote_files(url=url,flist=flist,dstdir=self.work)

    def auto_make(self,inc_dirs=[],args=[]):
        new_inc = [os.path.join(self.work,ii) for ii in inc_dirs]
        vpi_make(src_dirs=[self.work], inc_dirs=new_inc,args=args)
        os.system('mv *.vpi %s' % self.work)

    def custom_make(self,cmds=[]):
        cmdstr = ' && '.join(cmds)
        print(cmdstr)
        os.system(cmdstr)

class myhdl_vpi(custom_vpi):
    def __init__(self):
        tool='myhdl_vpi'
        url="https://raw.githubusercontent.com/myhdl/myhdl/master/cosimulation/icarus"
        flist = ['myhdl.c', 'myhdl_table.c', 'Makefile']
        custom_vpi.__init__(self,tool=tool,url=url,flist=flist) # , args=['-w'])

class fst_vpi(custom_vpi):
   
    def __init__(self):
        tool='fst_vpi'
        url="https://github.com/semify-eda/fstdumper/archive/refs/heads/main.zip"        
        self._init_work(tool=tool)
        cmds = ['cd %s' % os.path.join(self.work,'fstdumper-main'),
                'make simulation-iverilog',
                'mv *.vpi %s' % self.work ,
                'cd ..']
        custom_vpi.__init__(self,tool=tool,url=url,custom_make_cmds=cmds)

if __name__ == '__main__':
    # mv = myhdl_vpi()
    fv = fst_vpi()


#!/usr/bin/env python3

import os
from pueda.edalize import icarus
from pueda.common import get_clean_work
from pueda.vcd import vcd_view
import veriloggen as vg
from myhdl import *

def top2wrapper(topfile, topmodule,  filename='dut_wrapper.v', dump_en = False):

    modules = vg.from_verilog.read_verilog_module(topfile)
    top = modules[topmodule]   # dut real instance

    wrapper_name = os.path.basename(filename).split('.')[0]

    m = vg.Module(wrapper_name) # dut wrapper instance
    d = vg.StubModule(top.name) # dut stub instance
    dump = vg.StubModule('dump') # dumper module

    instlist = {}
    input_list = {}
    output_list = {}
    for io,iotype in top.io_variable.items():

        if isinstance(iotype,vg.core.vtypes.Input):            
            input_list[io] = instlist[io] = m.Reg(name=io, width=iotype.width)

        elif isinstance(iotype,vg.core.vtypes.Output):            
            output_list[io] = instlist[io] = m.Wire(name=io, width=iotype.width)
        else:
            assert(False)

    fromlist = input_list.values()
    # print(*fromlist)
    tolist = output_list.values()

    instlist['from_myhdl']  = m.Initial( vg.Systask('from_myhdl', *fromlist ) )
    instlist['to_myhdl']    = m.Initial( vg.Systask(  'to_myhdl', *tolist) )

    # copy paras and ports
    ports = {**input_list,**output_list}
    # print(ports)
    dut = m.Instance(d, 'dut', ports=ports)

    if dump_en:
        dump_i = m.Instance(dump, 'localdump')

    m.to_verilog(filename=filename)
    return wrapper_name

def top2signals(topfile='', topmodule=''):

    modules = vg.from_verilog.read_verilog_module(topfile)
    top = modules[topmodule]   # dut real instance

    slist = {}    
    for io,iotype in top.io_variable.items():

        if (isinstance(iotype,vg.core.vtypes.Input) or
            isinstance(iotype,vg.core.vtypes.Output)):
            slist[io] = Signal(intbv(0)[iotype.width:])

    return slist

def cosim_make_stub(topfile='',topmodule='',filename='dut_wrapper.v', dump_en=False):

    tool = 'myhdl'
    work = get_clean_work(tool=tool,makedir=True)
    full_file = os.path.join(work,filename)

    wrapper = top2wrapper(topfile=topfile, topmodule=topmodule, filename=full_file, dump_en=dump_en)

    return work, wrapper #, full_file

def cosimulation(vpi_path='./work_myhdl_vpi', vpi='myhdl', dut='', work='./work_icarus', ports={}):
    cmdstr = 'vvp -M %s -m%s %s' % (vpi_path, vpi, dut)
    
    os.chdir(work)
    print(cmdstr)
    return Cosimulation( cmdstr, **ports )

def myhdl_cosim_dut(topmodule='', topfile='',dump_en=False, ports={},
    simname ='', src_dirs=[],inc_dirs=[], stub_en = False):

    if stub_en:
        work, wrapper = cosim_make_stub(topfile=topfile,topmodule=topmodule, dump_en=dump_en)
    else:
        work = ''
        wrapper = topmodule

    # build dut
    print('##### Building icarus...')
    icarus_inst = icarus(
        simname=simname,
        top=wrapper,
        src_dirs = src_dirs + [work],
        inc_dirs = inc_dirs,
        dump_en=dump_en,
        myhdl_en = True,
        run_en=False)
    print('##### Building icarus... Done')

    # cosimulation
    return [ cosimulation(vpi_path=icarus_inst['mvpi'].work, 
                        work=icarus_inst['work_root'],
                        dut=simname, ports=ports),
            icarus_inst['work_root']]

@block
def clk_driver(clk, period=10):
    ''' Clock driver '''
    @always(delay(period//2))
    def driver():
        clk.next = not clk
    return driver

class myhdl_cosim_wrapper(object):
    topfile=''
    topmodule = ''
    simname=''
    src_dirs=[]
    inc_dirs=[]
    dump_en = True
    duration = 10    
    work = ''

    def __init__(self, topfile='', topmodule='', src_dirs=[], inc_dirs=[], simname='', dump_en = False, duration=200):
        self.topfile = topfile
        self.topmodule=topmodule
        self.simname = simname
        self.src_dirs = src_dirs
        self.inc_dirs = inc_dirs
        self.dump_en = dump_en
        self.duration = duration

        pass

    def dut_ports(self):
        return top2signals(topfile=self.topfile, topmodule=self.topmodule)

    def dut_instance(self, ports = {}):    
        dut, self.work = myhdl_cosim_dut( topfile=self.topfile, 
                                    topmodule=self.topmodule, 
                                    ports=ports,
                                    simname=self.simname,
                                    src_dirs=self.src_dirs,
                                    inc_dirs=self.inc_dirs, 
                                    dump_en=self.dump_en)
        return dut

    def sim_cfg(self,tb):
        self.sim = Simulation( tb )

    def sim_run(self,duration=0):
        self.sim.run(duration)

    def sim_view(self, vcd = 'dump.vcd', gtkw = '', mode='gtkwave'):
        if self.dump_en:
            vcd_view(os.path.join(self.work, fname = vcd, savefname = gtkw), block_en = False, mode=mode)



#!/usr/bin/env python3

import veriloggen as vg
from pueda.common import get_source_files_alldir

def sim(top='', src_dirs = [], inc_dirs = []):
    allfiles=get_source_files_alldir(src_dirs+inc_dirs)
    modules = vg.from_verilog.read_verilog_module(*allfiles,include=inc_dirs)
    pass

if __name__ == '__main__':
    sim()


#!/usr/bin/env python3

import sys
import os
sys.path.append( os.path.join(os.path.dirname(__file__), '../src') )

def test_verilator_counter():
    from pueda.edalize import verilator
    from pyverilator.verilator_tools import verilator_verilog_tb_ok

    if verilator_verilog_tb_ok():
        print('Verilator supports verilog testbench!')
        # tested with verilator 5.021
        verilator(
            simname='counter', top='counter_tb',                  
            src_dir = ['./counter/rtl','./counter/tb'],
            inc_dir = [],
            options=['--binary'],
            dump_en = False, sim_en=True)
    else:
        print('Verilator does not support verilog testbench!')
        # tested with verilator 4.099
        verilator(
            simname='counter', top='counter',
            src_dir = ['./counter/rtl','./counter/verilator'],
            inc_dir = [],
            dump_en = False, sim_en=True)
            # options=['--cc','--no-timing'],
        
if __name__ == '__main__':
    test_verilator_counter()            


#!/usr/bin/env python3

import os
import numpy as np
import shutil
from ast import literal_eval

# https://pypi.org/project/vcdvcd/
from vcdvcd import VCDVCD

def vcd_view(fname, mode='vcdterm', savefname='',options='', postcmd='', block_en=True):
    """ Visualize dump file, supports .vcd and .fst formats """
    assert os.path.isfile(fname), f'ERROR: file {fname} does not exist!'
    if mode == 'gtkwave':
        vcd_view_gtkwave(fname,savefname=savefname, options=options, postcmd=postcmd, block_en=block_en)
    elif mode == 'vcdterm':
        fname_wo_ext,ext=os.path.splitext(fname)
        if ext=='.fst':
            vcdfname = fname_wo_ext + '.vcd'
            os.system(f'fst2vcd {fname} -o {vcdfname}')            
            assert os.path.isfile(vcdfname), f'ERROR: file {vcdfname} does not exist!'
            fname = vcdfname
        vcd_view_terminal(fname, block_en=block_en)
    else:
        assert False, f'Unsupported view mode {mode} !!!'

def vcd_view_gtkwave(fname,savefname='',options='', postcmd='', block_en = True):

    # by default look for gtkw file named as vcd
    basefname = os.path.splitext(fname)[0]
    basegtkw = basefname + '.gtkw'
    
    if os.path.isfile(basegtkw):
        savefname = basegtkw

    if os.path.isfile(savefname):
        cmdstr = 'gtkwave %s -a %s %s %s' % (options, savefname,fname,postcmd)
    else:
        cmdstr = 'gtkwave %s %s %s' % (options,fname, postcmd)

    if not(block_en):
        cmdstr += ' &'

    # print(cmdstr)
    os.system(cmdstr)

def vcd_view_terminal(fname, block_en=False):
    """ view vcd in terminal """
    # check if vcd is installed
    path = shutil.which("vcd")
    assert not path is None, "vcd executable is not installed!!! \n\
        install with the following: \n\
        git clone https://github.com/yne/vcd.git \n\
        cd vcd \n\
        make \n\
        sudo make install"

    cmdstr = f'vcd < {fname}'
    if block_en:
        cmdstr += ' | less -S'
    else:
        cmdstr += ' | cat'

    print(cmdstr)
    os.system(cmdstr)

def vcd_tv2list(tv,fmt='dec'):
    """ converts .vcd file time-value lists, into a dictionary of time-value lists """
    time = []
    value = []
    for e in tv:

        if   fmt == 'dec':
            vstr = e[1]
        elif fmt == 'bin':
            vstr = '0b' + str.strip(e[1])
        elif fmt == 'hex':
            vstr = '0h' + str.strip(e[1])
        else:
            assert False, f'unsupported format {fmt} !!!'

        value.append(literal_eval(vstr))
        time.append(e[0])
    return {'value':value,'time':time}

def vcd_resample(s, fs=1):

    # generate time series
    tmin = min(s['time'])
    tmax = max(s['time'])
    t = np.arange(tmin,tmax,fs)
    sout = np.interp(t, s['time'], s['value'])
    # sout = f(t)

    return {'value':sout, 'time': t}    

def vcd_get_signal(vcdh, sname,fmt = 'dec', resample_en = False, fs=1e9):
    signal = vcdh[sname]
    # tv is a list of Time/Value delta pairs for this signal.
    s = vcd_tv2list(signal.tv,fmt = fmt)
    if resample_en:
        s = vcd_resample(s)
    return s


#!/usr/bin/python3

from vcd.gtkw import GTKWSave

def gen_gtkw_group(gtkw, gname, bit_signals, multibit_signals, nbits):

    with gtkw.group( gname ):
        multibit_signals_o = []
        for s in multibit_signals:
            multibit_signals_o.append( s + '[%d:0]' % (nbits-1) )

        signals = bit_signals + multibit_signals_o
        for s in signals:
            gtkw.trace( gname + s)

def gen_gtkw(fname = 'tb.gtkw', groups = [], nbits = 4):
    """
    groups = []

    # pwm mod
    groups.append(
        {
        'gname'            : 'tb.pwm.',
        'bit_signals'      : ['clk', 'resetn', 'pwm_o'],
        'multibit_signals' : ['inval', 'count']
        }
    )
    """

    fh = open(fname, "w")
    gtkw = GTKWSave(fh)

    for g in groups:
        gen_gtkw_group(gtkw, g['gname'], g['bit_signals'], g['multibit_signals'], nbits)

    pass



#!/usr/bin/env python3

import sys
import os
sys.path.append( os.path.join(os.path.dirname(__file__), '../src') )

from myhdl import *
from pueda.myhdl import myhdl_cosim_wrapper, clk_driver

@block
def tb(cosim):

    ports = cosim.dut_ports()
    dut_i = cosim.dut_instance(ports=ports)
    reset = ResetSignal(val=False,active=False,isasync=True)

    clk_driver_i = clk_driver(ports['clk'])
    clk = ports['clk']
    # reset = ports['reset']

    @always_seq(clk.posedge,reset=reset)
    def test_down_count():
        # or idx in range(15,0,-1):
        # print('INDEX: %d, COUNT %d' % (idx,ports['count']))
        # print('COUNT %d' % ports['count'])
        print('something')
        pass    

    return instances()

def myhdl_main():
    topfile = './counter/rtl/counter.v'
    cosim   = myhdl_cosim_wrapper(  topfile=topfile, topmodule='counter', 
                                    src_dirs=['./counter/rtl'], inc_dirs=[],
                                    simname='counter', dump_en=True)
        
    tb_i = tb(cosim)

    cosim.sim_cfg(tb_i)
    cosim.sim_run(16)
    # cosim.sim_view()

    pass

if __name__ == '__main__':
    myhdl_main()



#!/usr/bin/env python3

import os
import multiprocessing
import pkg_resources

# from edalize import *
from edalize.edatool import get_edatool # required by edalize 0.5.0

import pueda
from pueda.common import get_source_files_alldir, get_inc_list, get_clean_work
from pueda.icarus import myhdl_vpi, fst_vpi
from pueda.vcd    import vcd_view

def get_dump_dirs():
    r_data_path = '../../../../data/'
    # assert pkg_resources.resource_isdir(pueda.__name__,r_data_path), 'ERROR: directory {r_data_path} does not exist!'
    data_path = pkg_resources.resource_filename(pueda.__name__,r_data_path)

    if pkg_resources.resource_isdir(pueda.__name__,r_data_path):
        # pueda installed, look in shared data folder    
        inc_dirs = [ os.path.join(data_path, 'icarus/inc') ]
        src_dirs = [ os.path.join(data_path, 'icarus/src') ]
    else:
        # pueda not installed, assume running from local git copy
        inc_dirs = [ os.path.join(os.path.dirname(__file__), '../../data/icarus/inc') ]
        src_dirs = [ os.path.join(os.path.dirname(__file__), '../../data/icarus/src') ]

    return inc_dirs, src_dirs

def eda_get_files(dirlist,work_root,fmts=['.v','.sv','.vh'],print_en=False) -> list:
    fnames = get_source_files_alldir(dirlist,fmts=fmts)
    # print(fnames)
    
    # build list of dict as needed by edalize
    files = []
    for fname in fnames:

        if print_en:
            print(fname)

        fext = os.path.splitext(fname)[1]

        # source files
        if fext in ['.v']:
            f = {'name' : os.path.relpath(fname, work_root),
            'file_type' : 'verilogSource'}
        elif fext in ['.vh']:
            f = {'name' : os.path.relpath(fname, work_root),
            'file_type' : 'verilogSource',
            'is_include_file' : True
            }
        elif fext in ['.sv']:
            f = {'name' : os.path.relpath(fname, work_root),
            'file_type' : 'systemVerilogSource'}
        elif fext in ['.svh']:
            f = {'name' : os.path.relpath(fname, work_root),
            'file_type' : 'systemVerilogSource',
            'is_include_file' : True
            }
        elif fext in ['.c','.cpp']:
            f = {'name' : os.path.relpath(fname, work_root),
            'file_type' : 'cSource'}
        elif fext in ['.h']:
            f = {'name' : os.path.relpath(fname, work_root),
            'file_type' : 'cSource',
            'is_include_file' : True
            }
        elif fext in ['.vpi']:
            f = {'name' : os.path.relpath(fname, work_root),
            'file_type' : 'verilogSource'}
        else:
            print('unknown file extension for file %s !!!' % fname)
            f = {'name' : os.path.relpath(fname, work_root),
            'file_type' : 'unknown'}
          
        files.append(f)

    return files

def icarus(simname='', top='', src_dirs = [], inc_dirs = [],
            dump_en = True, dump_fst_vpi = True, run_en = True, myhdl_en = False,
            iverilog_options = [],
            plot_mode = 'vcdterm', plot_block_en = False, gtkw='') -> None:
    """ Icarus verilog helper function """

    # tool
    tool = 'icarus'
    work_root = get_clean_work(tool,True)

    if dump_en:
        inc_dump, src_dump = get_dump_dirs()
        inc_dirs += inc_dump
        src_dirs += src_dump

        iverilog_options += [
            '-DDUMP_EN', 
            '-DDUMP_LEVEL=0', 
            '-DDUMP_MODULE=%s' % top
            ]
        if dump_fst_vpi:
            iverilog_options += ['-DDUMP_FST_VPI']

    if myhdl_en:
        mvpi = myhdl_vpi()
        src_dirs += [mvpi.work]
    else:
        mvpi = None

    # this is only for fstdumper-vpi, but fst saving 
    # is enabled by default when using icarus with edalize
    if dump_en and dump_fst_vpi:
        fvpi = fst_vpi()
        vvp_options = ['-mfstdumper.so', '-M%s' % fvpi.work ]
    else:
        vvp_options = []

    # get design files
    files = eda_get_files(src_dirs, work_root, fmts=['.v','.vpi'])

    # get include directories
    options = iverilog_options + get_inc_list(inc_dirs,work_root)
    tool_options = {
        tool :
            {
            'iverilog_options'  : options,
            'vvp_options'       : vvp_options 
        }
    }

    edam = {
    'files'        : files,
    'name'         : simname,
    'toplevel'     : top,
    'tool_options' : tool_options
    }

    backend = get_edatool(tool)(edam=edam,
                                work_root=work_root)
    
    backend.configure()
    backend.build()

    if run_en:
        backend.run()
        if dump_en:
            if dump_fst_vpi:
                dump_file = 'dump.fst'
            else:
                dump_file = 'dump.vcd'

            vcd_view(os.path.join(work_root, dump_file),
                    savefname=gtkw,
                     mode=plot_mode, block_en=plot_block_en)

    return {'backend'   : backend,
            'work_root' : work_root, 
            'mvpi'      : mvpi
            }

def verilator(simname='', top='', src_dir=[], inc_dir = [],
              options = [],
              dump_en = True, dump_fst = False, gtkw = '', sim_en = True,
              plot_mode = 'vcdterm', plot_block_en = False) -> None:
    # tool
    tool = 'verilator'
    work_root = get_clean_work(tool)

    verilator_options = options
    verilator_options += [f'--top-module {top}' ]
    verilator_options += ['-j %d' % multiprocessing.cpu_count() ]

    if dump_en:
        # inc_dump, src_dump = get_dump_dirs()
        # inc_dir += inc_dump
        # src_dir += src_dump
    
        verilator_options += ['--trace']
            
        if dump_fst:
            verilator_options += ['--trace-fst', '-CFLAGS -DDUMP_FST']

    # get design files
    files = eda_get_files(src_dir, work_root, fmts=['.v','.vh','.cpp','.c'])

    # get include directories
    options = verilator_options + get_inc_list(inc_dir,work_root)
    tool_options = {
        tool :
            {
            'verilator_options'  : options,
        }
    }

    edam = {
    'files'        : files,
    'name'         : simname,
    'toplevel'     : top,
    'tool_options' : tool_options
    }

    backend = get_edatool(tool)(edam=edam,
                                work_root=work_root)

    os.makedirs(work_root)
    backend.configure()
    backend.build()

    if sim_en:
        backend.run()

        if dump_en:
            if dump_fst:
                dump_file = 'dump.fst'
            else:
                dump_file = 'dump.vcd'

            vcd_view( fname=os.path.join(work_root, dump_file),
                     savefname=gtkw,
                     options='-o', mode=plot_mode, block_en=plot_block_en)

def trellis(simname='',top='',src_dir=[], inc_dir=[]) -> None:

    # tool
    tool = 'trellis'
    work_root = get_clean_work(tool)
    
    # get design files
    files = eda_get_files(src_dir, work_root, fmts=['.v'])

    # get include directories
    files += eda_get_files(inc_dir, work_root, fmts=['.vh'])

    # get include directories
    options = get_inc_list(inc_dir,prefix='read -incdir ')

    tool_options = {
        tool :
            {
            'trellis_options'  : {'yosys_synth_options' : options},
        }
    }

    edam = {
    'files'        : files,
    'name'         : simname,
    'toplevel'     : top,
    'tool_options' : tool_options
    }

    backend = get_edatool(tool)(edam=edam,
                                work_root=work_root)

    os.makedirs(work_root)
    backend.configure()
    backend.build()
    backend.run()

def yosys_edalize(simname='',top='',src_dir=[], inc_dir=[], arch='ice40') -> None:

    # tool
    tool = 'yosys'
    work_root = get_clean_work(tool)
    
    # get design files
    files = eda_get_files(src_dir, work_root, fmts=['.v'])

    # get include directories
    files += eda_get_files(inc_dir, work_root, fmts=['.vh'])
    
    tool_options = {
        tool :
            {
            # 'yosys_synth_options'  : options,
            'arch': arch,
        },
        
    }

    edam = {
    'files'        : files,
    'name'         : simname,
    'toplevel'     : top,
    'tool_options' : tool_options
    }

    backend = get_edatool(tool)(edam=edam,
                                work_root=work_root)

    os.makedirs(work_root)
    backend.configure()
    backend.build()
    backend.run()


from setuptools import setup, find_packages
import os

datadir = os.path.join('data','icarus')
datafiles = [(d, [os.path.join(d,f) for f in files])
    for d, folders, files in os.walk(datadir)]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pueda',
    version='0.1.11',
    license='Apache 2.0',
    author="Marco Merlin",
    author_email='marcomerli@gmail.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    description="Collection of python for micro-Electronic Design Automation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/bat52/pueda',
    keywords='python EDA CAD',
    data_files = datafiles, 
    install_requires=[
          'wget',
          'edalize',
          'pyverilator-mm>=0.7.6',
          'dtrx',
          'myhdl',
          'veriloggen',
          'pyvcd',
          'vcdvcd',
          'numpy'
          ],
)



#!/usr/bin/python3

import os
import shutil

# load vcd_view for backwards compatibility
from pueda.vcd import vcd_view

def list2str(l,sep=' '):
    return sep.join(str(s) for s in l)

def get_source_files_alldir(dirlist,fmts=['.v','.sv','.vh'], excludes = []) -> list:
    files = []
    # create files list
    for d in dirlist:
        files = files + get_source_files(d,fmts=fmts, excludes=excludes)

    return files

def get_source_files(directory,fmts=['.v','.sv','.vh'], excludes = []) -> list:
    # print(directory)

    if os.path.isdir(directory):
        flist = os.listdir(directory)

        foutlist = []
        for f in flist:
            # print(f)
            fbase,fext = os.path.splitext(f)
            fullfile = os.path.abspath(os.path.join(directory,f))
            if os.path.isfile(fullfile) and (fext in fmts) and not(f in excludes) :
                # print('file %s' % f)
                foutlist.append(fullfile)
            elif os.path.isdir(fullfile):
                # recursively browse dir
                # print('recursion %s' % f)
                ldir = os.path.join(directory,f)
                # print(ldir)
                llist = get_source_files( ldir ,fmts=fmts )
                foutlist = foutlist + llist
            pass
    else: # if os.path.isdir(directory):
        print('%s : WARNING!!! directory "%s" does not exist!' % (__name__, directory) )
        foutlist = []

    # print(foutlist)

    return foutlist

def get_inc_list(inclist,work_root='',prefix='-I') -> list:

    outlist = []
    for ipath in inclist:
        if len(work_root)>0:
            inc = prefix + os.path.relpath(ipath,work_root)
        else:
            inc = prefix + os.path.abspath(ipath)
        outlist.append(inc)

    return outlist
    pass

def get_remote_files(url,flist=[],dstdir='./work_get'):
    import wget

    if not(os.path.isdir(dstdir)):
        os.makedirs(dstdir)

    if len(flist)>0:
        for f in flist:
            src = os.path.join(url   ,f)
            dst = os.path.join(dstdir,f)
            # print(src)
            # print(dst)
            if not(os.path.isfile(dst)):
                wget.download( src , dst )
    else:
        # single file
        wget.download( url , dstdir )
        fext = os.path.splitext(url)[1]
        if fext == '.zip':
            os.system('cd %s && dtrx -f `ls *.zip && cd ..`' % dstdir)

def get_work_path(tool=''):
    return os.path.join(os.getcwd() , 'work_' + tool)    

def work_exists(tool=''):
    return os.path.isdir(get_work_path(tool))

def get_clean_work(tool='',makedir=False,rm_en=True):
    work_root = get_work_path(tool) 

    if work_exists(tool) and rm_en:
        # delete work directory
        shutil.rmtree(work_root,ignore_errors=True)
    
    if makedir:
        os.makedirs(work_root)
        
    return work_root

def write_file_lines(fname, lines=[], mode='w', print_en=False):
    
    f = open(fname, mode)
    for l in lines: 
        # print(l)
        f.write(l + '\n')
    f.close()

    # print file
    if print_en:
        os.system('cat %s' % fname)


#!/usr/bin/env python3

import os
import shutil
import pyverilator
from pueda.vcd import vcd_view

class pyverilator_wrapper(object):
    ''' generate a dut wrapper with pyverilator '''
    sim = None
    dump_filename = ''

    def __init__(self, fname='', src_dirs=[], command_args = [],
                 dump_en = False, dump_fst = False, dump_filename = 'dump', dump_level=0):
        ''' initialize pyverilator wrapper '''

        # rename to .v, if .sv
        if not os.path.isfile(fname):
            assert False, f'File {fname} does not exist!'

        base,ext = os.path.splitext(fname)
        # print(ext)

        if ext == '.sv':
            print('renaming input file to .v')
            ofname = base + '.v'
            shutil.copyfile(fname, ofname)
        else:
            ofname = fname
        print(ofname)

        self.sim = pyverilator.PyVerilator.build(ofname,
                                                 verilog_path=src_dirs,
                                                 args=command_args,
                                                 dump_fst=dump_fst,
                                                 dump_level=dump_level)
        if dump_en:
            if dump_fst:
                dump_ext = '.fst'
            else:
                dump_ext = '.vcd'
            self.dump_filename = dump_filename + dump_ext
            # start gtkwave to view the waveforms as they are made
            self.sim.start_vcd_trace(self.dump_filename)
            # self.view_waves()

    def view_waves(self,savefname='',options='', postcmd='', block_en = True, mode='gtkwave'):
        ''' view vcd waves with gtkwave '''

        # if False:
        #    # start gtkwave to view the waveforms as they are made
        #    self.sim.start_gtkwave()
        #
        #    # add all the io and internal signals to gtkwave
        #    # self.sim.send_to_gtkwave(self.sim.io)
        #    # self.sim.send_to_gtkwave(self.sim.internals)
        # else:
        vcd_view(self.dump_filename,
                    savefname=savefname,
                    options=options,
                    postcmd=postcmd,
                    block_en=block_en,
                    mode=mode)
